
!--------*---------*---------*---------*---------*---------*---------*-*
!     
! Subroutine: eigen_s
!
! Purpose
! =======
!
! implementation of interface eigen_s (see eigen_lib.F module)
!
!
! Arguments
! =========
!
! n, nvec, lda, ldz    (input) integer
!  lda and ldz are the leading dimensions corresponding to a and z,
!  respectively. These values must be equal or larger than one obtained
!  by calling eigen_get_matdims.
!
! m_forward, m_backward  (input) integer, optional
!  default values are m_forward = 48, and m_backward = 128.
!
! mode  (input) character, optional
!  = 'A' : eigenvalues and corresponding eigenvectors (default)
!  = 'N' : eigenvalues only
!  = 'X' : mode 'A' + improve accuracy of eigenvalues
!
!--------*---------*---------*---------*---------*---------*---------*-*

      subroutine eigen_s0(n, nvec, a, lda, w, z, ldz,
     &     m_forward, m_backward, mode)

      use eigen_scaling_mod
      use trbakwy4_mod
      use eigen_trd_mod
      use bisect_mod
      use dc2_mod
      !
      ! Trick to avoid a conflict between func-name and interface
      !
      use eigen_libs_mod, eigen_s0_interface => eigen_s0
      use eigen_devel_mod
!      use comm_mod, only : barrier
      use mpi, only : MPI_COMM_NULL
!$    use omp_lib

      implicit none

      integer,   intent(in)           :: n
      integer,   intent(in), optional :: nvec
      real(8),   intent(inout)        :: a(lda,*)
      integer,   intent(in)           :: lda
      real(8),   intent(out)          :: w(*)
      real(8),   intent(out)          :: z(ldz,*)
      integer,   intent(in)           :: ldz
      integer,   intent(in), optional :: m_forward
      integer,   intent(in), optional :: m_backward
      character, intent(in), optional :: mode

      real(8), allocatable   :: d(:), e(:), e2(:)

      integer                :: nvec_, m_f, m_b
      character              :: mode_

      real(8)                :: SIGMA
      integer                :: nm, ny, nb
      integer                :: my_rank, world_size, ierr
      integer                :: INFO
      logical                :: flag

      real(8)                :: hs0, hs1, s0, s1
      real(8)                :: ret, ret_1, ret_2, ret_3
!-
!-------------------------------------------------------------------
!-
      call eigen_get_initialized(flag)
      if (.not. flag) then
        return
      end if
      if (TRD_COMM_WORLD == MPI_COMM_NULL) then
        return
      end if
!-
!-------------------------------------------------------------------
!-
      if (n <= 0) then
        print*,"Warining: Negative dimesion is invalid!"
        return
      end if
!-
!-------------------------------------------------------------------
!-
      if (present(nvec)) then
        nvec_ = nvec
      else
        nvec_ = n
      end if

      if (present(mode)) then
        mode_ = mode
      else
        mode_ = 'A'
      end if

      if (present(m_forward)) then
        m_f = m_forward
      else
        m_f = eigen_NB_f
      end if

      if (present(m_backward)) then
        m_b = m_backward
      else
        m_b = eigen_NB_b
      end if
!-
!-------------------------------------------------------------------
!-
      call eigen_get_matdims0(n, nm, ny, m_f, m_b)
      if (nm <= 0 .or. ny <= 0) then
        print*,"Warining: Problem size is too large for"
     &       // " 32bit fortarn integer binary."
        return
      end if
!-
!-------------------------------------------------------------------
!-
      hs0 = eigen_get_wtime()
      ret_1 = ZERO; ret_2 = ZERO; ret_3 = ZERO
!-
!-------------------------------------------------------------------
!-
      allocate( d(1:n), e(1:n), stat=ierr )
      if ( ierr /= 0 ) then
        call eigen_abort( "Memory allocation error [eigen_s].", 1 )
      end if
#if DEBUG
      if ( TRD_inod == 1 ) print*,"global data allocation done"
#endif

      world_size = TRD_nnod
      my_rank    = TRD_inod-1
!-
!-------------------------------------------------------------------
!-
      SIGMA = ONE
      call eigen_scaling(n, a(1,1), lda, SIGMA)
!-
!-------------------------------------------------------------------
!-
#if TIMER_PRINT
      s0 = eigen_get_wtime()
#endif

#if DEBUG
      if ( TRD_inod == 1 ) print*,"eigen_trd is going to be invoked"
#endif

      if (mode_ == 'N' .OR. nvec_ <= 0) then
        call eigen_trd(n, a(1,1), lda, d(1), e(1), m_f)
      else
        call eigen_trd(n, a(1,1), lda, w(1), e(1), m_f)
      end if
      ret_1 = dble(n)**3*4/3

#if TIMER_PRINT
      s1 = eigen_get_wtime()
      if (my_rank==0) then
        print 10000,"TRD-BLK ", n, s1-s0, 1D-9*ret_1/(s1-s0), "GFLOPS"
      end if
      call flush(6)
#endif
!-
!-------------------------------------------------------------------
!-
#if TIMER_PRINT
      s0 = eigen_get_wtime()
#endif
#if DEBUG
      if ( TRD_inod == 1 ) print*,"eigen_dc2 is going to be invoked"
#endif

      if ((mode_ == 'A' .or. mode_ == 'X') .and. nvec_ > 0) then
        if (mode_ == 'X') then
          allocate( e2(1:n), stat=ierr )
          if ( ierr /= 0 ) then
            call eigen_abort( "Memory allocation error [eigen_s].", 1 )
          end if
          d (1:n) = w(1:n)
          e2(1:n) = e(1:n)
          call eigen_dc2(n, nvec_, w, e2(2), z, ldz, INFO, ret_2)
          call eigen_bisect(d, e, w, n, 1)
          deallocate( e2 )
        else
          d (1:n) = e(1:n)
          call eigen_dc2(n, nvec_, w, d(2), z, ldz, INFO, ret_2)
        end if
      end if
      if (mode_ == 'S') then
        call eigen_identity(n, z, ldz)
        d(1:n) = w(1:n)
        call eigen_bisect(d, e, w, n, 0)
      else if (mode_ == 'N' .OR. nvec_ <= 0) then
        call eigen_bisect(d, e, w, n, 0)
      end if
      deallocate( d )

#if TIMER_PRINT
      s1 = eigen_get_wtime()
      if (my_rank == 0) then
        print 10000,"D&C     ", n, s1-s0, 1D-9*ret_2/(s1-s0), "GFLOPS"
      end if
      call flush(6)
#endif
      if (mode_ == 'N' .OR. nvec_ <= 0) then
        goto 99999
      end if
!-
!-------------------------------------------------------------------
!-
#if TIMER_PRINT
      s0 = eigen_get_wtime()
#endif
#if DEBUG
      if ( TRD_inod == 1 ) print*,"eigen_trbakwy will be invoked"
#endif

      nb = 1
      call eigen_common_trbakwy(n, nvec_, a(1,1), lda, z(1,1), ldz,
     &     e, m_b, nb)
      ret_3 = 2*dble(nvec_)*dble(n)**2

#if TIMER_PRINT
      s1 = eigen_get_wtime()
      if (my_rank == 0) then
        print 10000,"TRDBAK  ", n, s1-s0, 1D-9*ret_3/(s1-s0), "GFLOPS"
      end if
      call flush(6)
#endif
!-
!-------------------------------------------------------------------
!-
      if (SIGMA /= ONE) then
        SIGMA = ONE / SIGMA
        call dscal(n, w, SIGMA, 1)
      end if
!-
!-------------------------------------------------------------------
!-
#if TIMER_PRINT
      hs1 = eigen_get_wtime()
      ret = ret_1 + ret_2 + ret_3
      if (my_rank == 0) then
        print*,"Total(eigen_s)",
     &       hs1-hs0, 1D-9*ret/(hs1-hs0), "GFLOPS"
      end if
      call flush(6)
#endif
!-
!-------------------------------------------------------------------
!-
99999 continue
!-
!-------------------------------------------------------------------
!-
      hs1 = eigen_get_wtime()
      ret = ret_1 + ret_2 + ret_3
      if (ret_2 == 0) ret = -ret
      a(1, 1) = ret
      a(2, 1) = hs1-hs0
#if TIMER_PRINT
      a(3, 1) = comm_time_reduction
     &     + comm_time_dc
     &     + comm_time_backtrafo
#else
      a(3, 1) = -1d0
#endif
!-
!-------------------------------------------------------------------
!-
      deallocate( e )
!-
!-------------------------------------------------------------------
!-
10000 format (X, A8, I8, E25.16e2, E25.16e2, X, A)
!-
!-------------------------------------------------------------------
!-
      end subroutine eigen_s0
